/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "constraint.h"
#include "encoding.h"
#include "funcdef.h"

constraint::constraint()
{
	refcount = 1;
	treewalk_constraintid = -1;
}

constraint::~constraint()
{
	assert(refcount==0);
}

constraint *constraint::parse(const encoding *e,char *&s)
{
	uint32_t i;
	int len = 0;
	if(isdigit(*s))
	{
		/* constant field */
		uint32_t val;
		if((sscanf(s,"0x%x%n",&val,&len) != 1) && (sscanf(s,"%u%n",&val,&len) != 1))
			return NULL;
		s += len;
		return new constraint_const(val);
	}
	if(!isalpha(*s) && (*s != '_'))
		return NULL;
	/* field or func */
	uint32_t startbit,endbit;
	if(sscanf(s,"opcode<%u>%n",&startbit,&len) == 1) /* pseudofield */
	{
		if(startbit > opcodesize)
			return NULL;
		s += len;
		return new constraint_field(startbit,1);
	}
	if(sscanf(s,"opcode<%u:%u>%n",&endbit,&startbit,&len) == 2)
	{
		if((startbit > opcodesize) || (endbit < startbit))
			return NULL;
		s += len;
		return new constraint_field(startbit,(endbit-startbit)+1);
	}
	while(isalnum(s[len]) || (s[len] == '_'))
		len++;
	char temp = s[len];
	s[len] = 0;
	if(temp != '(')
	{
		for(i=0;i<e->numfields;i++)
		{
			if(!strcmp(e->fields[i].name,s))
			{
				s[len] = temp;
				s += len;
				return new constraint_field(&e->fields[i]);
			}
		}
		/* Not a field name! */
		s[len] = temp;
		return NULL;
	}
	/* it's a function */
	funcdef *f = funcdef_get(s);
	s[len] = temp;
	s += len+1;
	if(f == NULL)
		return NULL;
	constraint **c = new constraint *[f->getnumparam()];
	assert(c);
	memset(c,0,sizeof(constraint *)*f->getnumparam());
	bool numeric=false;
	bool identical=true;
	for(i=0;i<f->getnumparam();i++)
	{
		while(*s && isspace(*s))
			s++;
		if(i)
		{
			if(*(s++) != ',')
				goto err;
			while(*s && isspace(*s))
				s++;
		}
		c[i] = constraint::parse(e,s);
		if(!c[i])
			goto err;
		bool numeric2;
		if(c[i]->gettype() == EFUNC)
			numeric2 = static_cast<constraint_func *>(c[i])->getfunc()->numeric_output;
		else
			numeric2 = true;
		if(i)
			identical &= (numeric == numeric2);
		numeric = numeric2;
	}
	while(*s && isspace(*s))
		s++;
	if(*(s++) != ')')
		goto err;
	/* Type safety checks */
	if(f->same_input && !identical)
		goto err; /* TODO - Sort out some proper error reporting! */
	if(f->numeric_input && !numeric)
		goto err;
	return new constraint_func(f,c);
err:
	while(i--)
		dereference(c[i]);
	delete c;
	return NULL;
}

bool constraint::deduplicate(constraint **a,constraint **b)
{
	if((*a == *b) || !(**a == **b))
		return false;
	/* Merge into *a */
	dereference(*b);
	*b = *a;
	(*a)->refcount++;
	return true;
}

bool constraint::dereference(constraint *c)
{
	assert(c);
	if(--(c->refcount))
		return false;
	delete c;
	return true;
}

void constraint::reference()
{
	refcount++;
}

void constraint::treewalk_number(cntype type,int *count,vector<constraint *> &constraints)
{
	if((type == gettype()) && (treewalk_constraintid == -1))
	{
		treewalk_constraintid = (*count)++;
		constraints.push_back(this);
	}
}

constraint_field::constraint_field(const field *f)
{
	assert(f);
	startbit = f->startbit;
	bitcount = f->bitcount;
}

constraint_field::constraint_field(uint32_t start,uint32_t count)
{
	startbit = start;
	bitcount = count;
}

constraint_field::~constraint_field()
{
	/* Nothing */
}

cntype constraint_field::gettype() const
{
	return EFIELD;
}

uint32_t constraint_field::eval(uint32_t opcode) const
{
	return (opcode >> startbit) & ((((uint32_t)1)<<bitcount)-1);
}

void constraint_field::emit(const char *opcode) const
{
	fprintf(fout,"((%s >> %d) & 0x%x)",opcode,startbit,((((uint32_t)1)<<bitcount)-1));
}

void constraint_field::emit_funcs() const
{
}

bool constraint_field::operator==(const constraint &c) const
{
	if(c.gettype() != EFIELD)
		return false;
	const constraint_field *f = static_cast<const constraint_field *>(&c);
	return (f->startbit == startbit) && (f->bitcount == bitcount);
}

bool constraint_field::equals(const constraint &c) const
{
	if(c.gettype() != EFIELD)
		return false;
	const constraint_field *f = static_cast<const constraint_field *>(&c);
	return (f->startbit == startbit) && (f->bitcount == bitcount);
}

uint32_t constraint_field::getfields() const
{
	return ((((uint32_t)1)<<bitcount)-1)<<startbit;
}

void constraint_field::collect(vector<constraint **> &c) const
{
}

string constraint_field::describe(const field *fields,uint32_t numfield) const
{
	/* Try and find the field in the list we've been given */
	while(numfield--)
		if((fields[numfield].startbit == startbit) && (fields[numfield].bitcount == bitcount))
			return string(fields[numfield].name);
	/* Else generate our own name
	   Use ARM-ish syntax of opcode<hibit:lobit> */
	char buf[32];
	if(bitcount>1)
		sprintf(buf,"opcode<%d:%d>",startbit+bitcount-1,startbit);
	else
		sprintf(buf,"opcode<%d>",startbit);
	return string(buf);
}

uint32_t constraint_field::getstartbit() const
{
	return startbit;
}

uint32_t constraint_field::getbitcount() const
{
	return bitcount;
}

constraint_const::constraint_const(uint32_t v)
{
	val = v;
}

constraint_const::~constraint_const()
{
	/* Nothing */
}

cntype constraint_const::gettype() const
{
	return ECONST;
}

uint32_t constraint_const::eval(uint32_t opcode) const
{
	return val;
}

void constraint_const::emit(const char *opcode) const
{
	fprintf(fout,"0x%x",val);
}

void constraint_const::emit_funcs() const
{
}

bool constraint_const::operator==(const constraint &c) const
{
	if(c.gettype() != ECONST)
		return false;
	const constraint_const *co = static_cast<const constraint_const *>(&c);
	return (co->val == val);
}

bool constraint_const::equals(const constraint &c) const
{
	if(c.gettype() != ECONST)
		return false;
	const constraint_const *co = static_cast<const constraint_const *>(&c);
	return (co->val == val);
}

uint32_t constraint_const::getfields() const
{
	return 0;
}

void constraint_const::collect(vector<constraint **> &c) const
{
}

string constraint_const::describe(const field *fields,uint32_t numfield) const
{
	char buf[32];
	sprintf(buf,"0x%x",val);
	return string(buf);
}

uint32_t constraint_const::getvalue() const
{
	return val;
}

constraint_func::constraint_func(funcdef *f,constraint **p)
{
	assert(f);
	assert(p);
	func = f;
	params = p;
	fields = 0;
	for(uint32_t i=0;i<f->getnumparam();f++)
		fields |= params[i]->getfields();
}

constraint_func::~constraint_func()
{
	for(uint32_t i=0;i<func->getnumparam();i++)
		dereference(params[i]);
	delete[] params;
}

cntype constraint_func::gettype() const
{
	return EFUNC;
}

uint32_t constraint_func::eval(uint32_t opcode) const
{
	if(func == funcdef_land)
	{
		/* Lazy evaluation optimisation */
		return params[0]->eval(opcode) && params[1]->eval(opcode);
	}
	else if(func == funcdef_lor)
	{
		/* Lazy evaluation optimisation */
		return params[0]->eval(opcode) || params[1]->eval(opcode);
	}
	uint32_t args[MAX_FUNC_PARAMS];
	for(uint32_t i=0;i<func->getnumparam();i++)
		args[i] = params[i]->eval(opcode);
	return func->eval(args);
}

void constraint_func::emit(const char *opcode) const
{
	fprintf(fout,"%sfunc_%s(",prefix,func->getname());
	for(uint32_t i=0;i<func->getnumparam();i++)
	{
		if(i)
			putc(',',fout);
		params[i]->emit(opcode);
	}
	putc(')',fout);
}

void constraint_func::emit_funcs() const
{
	func->emit();
	for(uint32_t i=0;i<func->getnumparam();i++)
		params[i]->emit_funcs();
}

bool constraint_func::operator==(const constraint &c) const
{
	if(c.gettype() != EFUNC)
		return false;
	const constraint_func *f = static_cast<const constraint_func *>(&c);
	if(f->func != func)
		return false;
	for(uint32_t i=0;i<func->getnumparam();i++)
		if(params[i] != f->params[i])
			return false;
	return true;
}

bool constraint_func::equals(const constraint &c) const
{
	if(c.gettype() != EFUNC)
		return false;
	const constraint_func *f = static_cast<const constraint_func *>(&c);
	if(f->func != func)
		return false;
	for(uint32_t i=0;i<func->getnumparam();i++)
		if(!params[i]->equals(*f->params[i]))
			return false;
	return true;
}

uint32_t constraint_func::getfields() const
{
	uint32_t fields=0;
	for(uint32_t i=0;i<func->getnumparam();i++)
		fields |= params[i]->getfields();
	return fields;
}

void constraint_func::collect(vector<constraint **> &c) const
{
	for(uint32_t i=0;i<func->getnumparam();i++)
	{
		size_t j;
		for(j=0;j<c.size();j++)
			if(*c[j] == params[i])
				break;
		if(j == c.size())
		{
			params[i]->collect(c);
			c.push_back(&params[i]);
		}
	}
}

string constraint_func::describe(const field *fields,uint32_t numfield) const
{
	string c(func->getname());
	for(uint32_t i=0;i<func->getnumparam();i++)
	{
		c += (i?",":"(");
		c += params[i]->describe(fields,numfield);
	}
	c += ")";
	return c;
}

funcdef *constraint_func::getfunc() const
{
	return func;
}

constraint **constraint_func::getparams() const
{
	return params;
}

void constraint_func::treewalk_number(cntype type,int *count,vector<constraint *> &constraints)
{
	constraint::treewalk_number(type,count,constraints);
	for(uint32_t i=0;i<func->getnumparam();i++)
		params[i]->treewalk_number(type,count,constraints);
}

void simplify_and_insert(encoding *e,constraint *c)
{
	/* We want to simplify:
	   lnot(lnot(X)) -> X
	   land(X,Y) -> split in two
	   lnot(lor(X,Y)) -> split into lnot(X), lnot(Y)
	*/
	if(c->gettype() == EFUNC)
	{
		constraint_func *f = static_cast<constraint_func *>(c);
		if((f->getfunc() == funcdef_lnot) && (f->getparams()[0]->gettype() == EFUNC))
		{
			constraint_func *f2 = static_cast<constraint_func *>(f->getparams()[0]);
			if(f2->getfunc() == funcdef_lnot)
			{
				simplify_and_insert(e,f2->getparams()[0]);
				return;
			}
			else if(f2->getfunc() == funcdef_lor)
			{
				constraint **d = new constraint *[1];
				d[0] = f2->getparams()[0];
				d[0]->reference();
				constraint *n = new constraint_func(funcdef_lnot,d);
				simplify_and_insert(e,n);
				constraint::dereference(n);

				d = new constraint *[1];
				d[0] = f2->getparams()[1];
				d[0]->reference();
				n = new constraint_func(funcdef_lnot,d);
				simplify_and_insert(e,n);
				constraint::dereference(n);

				return;
			}
		}
		else if(f->getfunc() == funcdef_land)
		{
			simplify_and_insert(e,f->getparams()[0]);
			simplify_and_insert(e,f->getparams()[1]);
			return;
		}
	}
	e->constraints = (constraint **) realloc(e->constraints,sizeof(constraint *)*(e->numconstraints+1));
	assert(e->constraints);
	e->constraints[e->numconstraints++] = c;
	c->reference();
}
