/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/
#ifndef NODE_H
#define NODE_H

#include "common.h"
#include "opcodesubset.h"

class node_leaf;

class opt_strings_action
{
public:
	opt_strings_action(const encoding *e);
	~opt_strings_action();

	tokenset code;
	vector<string> args;

	void emit(node_leaf *parent);
	string get_name() const;

private:
	const encoding *enc;
	bool emitted;
};

typedef enum {
//	EARG_STRING,
	EARG_FIELD,
	EARG_NONSTANDARD,
	EARG_ARGSTRUCT,
} argtype;

typedef struct {
	argtype type;
	string str; /* Value of string argument, or local variable name (i.e. field name or 'nonstandard') if leaf node */
	uint32_t shouldmatch_value,shouldmatch_mask; /* For EARG_NONSTANDARD */
	vector<field *> fields; /* For EARG_FIELD, ordered from low component to high */
} argument;

typedef enum {
	ECONSTRAINT, /* Node checks a constraint */
	EMASK, /* Node checks a mask */
	ELEAF, /* Leaf node which performs an action */
} ntype;

/* Helper struct for node::collect */
typedef struct {
	node **n;
	int i;
} collecti;

class node
{
private:
	uint32_t refcount;
protected:
	uint32_t numchildren;
	const encodingset *results;
	node **children;
	opcode_subset *childset;

	/* Regular code generation state */
	bool emitted_prototype;
	bool emitted_body;
	bool emitted_graph;
	bool prepared_args;

	node(uint32_t num,const opcode_subset *o,const encodingset *res);

	virtual ~node();

	void collect_int(vector<collecti> &c) const;
public:
	virtual ntype gettype() const = 0;

	virtual void emit_prototype(); /* Emit function prototype (if required) */
	virtual void emit_args() const;
	virtual void emit_locals() const;

	virtual void emit_call() const; /* Emit call to subnode (or perform action if leaf node) */

	virtual void emit_body() = 0; /* Emit function body (if required) */

	virtual const encoding *verify(uint32_t opcode) const = 0; /* Ask the tree what it thinks an opcode decodes to, for verifying tree correctness */

	virtual void emit_graph() = 0; /* Emit graphml elements */

	virtual string get_name() const = 0; /* Return node name suitable for function names, etc. */

	virtual void prepare_args(); /* Prepare argument list for this node + children */

	void treewalk_number(ntype type,int *count,vector<node *> &nodes); /* Number things for the tree walker, filling in the given node list in order of assignment */

	uint32_t getnumchildren() const; /* Return number of child nodes required */

	const encodingset &getchildset(uint32_t i) const; /* Return set of encodings which the child needs to consider */

	const opcode_subset *getsubset(uint32_t i) const; /* Return subset which is already decoded */

	const node *getchildnode(uint32_t i) const; /* Return child node */

	node *getchildnode(uint32_t i); /* Return child node */

	void setchildren(node **arr); /* Set children */

	void collect(vector<node **> &n) const; /* Collect pointers to all the child node pointers for easy deduplication */
	void collect(vector<node *> &n) const; /* Collect pointers to all the child nodes */

	static bool deduplicate(node **a,node **b); /* returns true if b gets deduplicated into a */
	static bool dereference(node *d); /* returns true when object gets deleted */
	void reference();

	virtual bool operator==(const node &n) const = 0;

	void calcdepth(uint32_t &min,uint32_t &max); /* Calculate min/max depth of tree */

	static void shrink(node *&n); /* Shrink tree by eliminating redundant nodes */

	virtual bool sametest(const node &n) const = 0; /* Like ==, but doesn't check if children are identical */

	map<string,argument> function_args;
	map<string,argument> local_vars;

	/* Compact tree walker generation state */
	int treewalk_nodeid;
};

class node_constraint : public node
{
private:
	const constraint *con;
	opcode_subset subset;

	~node_constraint();
public:
	node_constraint(const constraint *c,const opcode_subset *o,const encodingset *res);

	ntype gettype() const;
	void emit_prototype();
	void emit_body();
	void emit_graph();
	string get_name() const;
	const encoding *verify(uint32_t opcode) const;
	bool operator==(const node &n) const;
	bool sametest(const node &n) const;

	const constraint *getconstraint() const;
};

class node_mask : public node
{
private:
	uint8_t startbit;
	uint8_t bitcount;
	opcode_subset subset;

	~node_mask();
public:
	node_mask(uint32_t start,uint32_t count,const opcode_subset *o,const encodingset *res);

	ntype gettype() const;
	void emit_body();
	void emit_graph();
	string get_name() const;
	bool sametest(const node &n) const;
	const encoding *verify(uint32_t opcode) const;
	bool operator==(const node &n) const;

	void shrink2();
	uint8_t getstartbit() const;
	uint8_t getbitcount() const;
};

class node_leaf : public node
{
private:
	const encoding *enc; /* Root kin which this node matches */
	opt_strings_action *opt_action; /* Optimised action to perform instead of original */
	map<string,string> opt_action_args;

	~node_leaf();
public:
	node_leaf(const encoding *e);

	ntype gettype() const;
	void emit_body();
	void emit_graph();
	string get_name() const;
	const encoding *verify(uint32_t opcode) const;
	bool operator==(const node &n) const;
	bool sametest(const node &n) const;
	void prepare_args();

	const encoding *get_encoding() const;
	void set_opt_strings_action(opt_strings_action *a,map<string,string> &args);
};


#endif
